/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.applet;

import net.hizlab.kagetaka.Reporter;
import net.hizlab.kagetaka.Resource;
import net.hizlab.kagetaka.build.ParserManager;
import net.hizlab.kagetaka.cookie.CookieStore;
import net.hizlab.kagetaka.rendering.Document;
import net.hizlab.kagetaka.rendering.EngineListener;
import net.hizlab.kagetaka.rendering.HawkContext;
import net.hizlab.kagetaka.rendering.HawkEngine;
import net.hizlab.kagetaka.rendering.HawkPanel;
import net.hizlab.kagetaka.rendering.Option;
import net.hizlab.kagetaka.rendering.PanelListener;
import net.hizlab.kagetaka.rendering.Request;
import net.hizlab.kagetaka.rendering.RequestList;
import net.hizlab.kagetaka.token.FormItem;
import net.hizlab.kagetaka.util.Environment;
import net.hizlab.kagetaka.util.StringUtils;

import java.applet.Applet;
import java.awt.BorderLayout;
import java.awt.Cursor;
import java.awt.Dimension;
import java.awt.Color;
import java.awt.Font;
import java.awt.Graphics;
import java.awt.Image;
import java.awt.MediaTracker;
import java.awt.MenuItem;
import java.awt.Point;
import java.awt.TextField;
import java.awt.Toolkit;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.image.ImageProducer;
import java.io.IOException;
import java.net.HttpURLConnection;
import java.net.URL;
import java.net.URLConnection;
import java.net.MalformedURLException;
import java.util.Locale;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;

/**
 * ΥץåȤǤ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.15 $
 */
public class HawkApplet
	extends Applet
{
	private static final String RESOURCE = "net.hizlab.kagetaka.applet.Resource";
	private static final String BUILD_RESOURCE = "net.hizlab.kagetaka.Build";
	static {
		Environment.isApplet = true;
	}
	
	/** @serial ݡȥץȥ */
	private Hashtable supportProtocol = new Hashtable();
	{
		StringTokenizer st = new StringTokenizer(Resource.getMessage("config.support.protocol", null), ",; ");
		while (st.hasMoreTokens())
			supportProtocol.put(st.nextToken(), "");
	}
	
	/** @serial 륳ƥ */
	private HawkAppletContext hawkContext;
	/** @serial ɥ쥹С */
	private TextField         addressBar;
	/** @serial ơСʸ */
	private String            statusText;
	/** @serial ץ */
	private Option            option     = new Option();
	/** @serial ٥ */
	private int               logLevel   = Reporter.ERROR;
	/** @serial ˬѤ URL */
	private Hashtable         visitedUrl = new Hashtable();
	/** @serial åޥ͡ */
	private CookieStore       cookieManager;
	
	/**
	 * 륢ץåȤΥ󥹥󥹤ޤ
	 */
	public HawkApplet()
	{
	}
	
	/**
	 * ץåȤޤ
	 */
	public void init() {
		this.addressBar = new TextField();
		addressBar.setEditable(false);
		
		String value = null;
		boolean useAddressBar = true;
		
		// եȤ
		if ((value = getParameter("font")) != null) {
			Font font = Font.decode(value);
			if (font != null)
				option.setDefaultFont(font);
			else
				showStatus(getMessage("parameter.nofont", new String[]{value}));
		}
		
		// ѡ
		if ((value = getParameter("parser")) != null) {
			if (ParserManager.getParser(value) != null)
				option.setInputStreamParser(value);
			else
				showStatus(getMessage("parameter.noparser", new String[]{value}));
		}
		
		// ɥ쥹С
		if ((value = getParameter("addressbar")) != null)
			useAddressBar = Boolean.valueOf(value).booleanValue();
		
		// ɥ쥹С
		if ((value = getParameter("scrollbar")) != null)
			option.setShowScrollbar(Boolean.valueOf(value).booleanValue());
		
		// GIF ž
		if ((value = getParameter("spingif")) != null)
			option.setSpinGifImage(Boolean.valueOf(value).booleanValue());
		
		// JPEG ž
		if ((value = getParameter("spinjpeg")) != null)
			option.setSpinJpegImage(Boolean.valueOf(value).booleanValue());
		
		// ƥȽ饤
		if ((value = getParameter("swapdecoration")) != null)
			option.setSwapDecoration(Boolean.valueOf(value).booleanValue());
		
		// LogLevel 
		if ((value = getParameter("loglevel")) != null)
			logLevel = Integer.parseInt(value);
		
		// URL 
		if ((value = getParameter("url")) == null) {
			showStatus(getMessage("parameter.nourl", null));
			return;
		}
		
		URL url = null;
		try {
			url = new URL(getCodeBase(), value);
		} catch (MalformedURLException e) {
			showStatus(getMessage("error.url.unknown", new String[]{value, e.toString()}));
			return;
		}
		
		// 
		this.cookieManager = CookieStore.getInstance();
		
		this.hawkContext = new HawkAppletContext();
		
		// ̤
		setLayout(new BorderLayout());
		if (useAddressBar)
			add(addressBar, BorderLayout.NORTH );
		add(hawkContext.hawkPanel, BorderLayout.CENTER);
		
		// եɤ߹
		open(url);
	}
	
	/**
	 * ٤ƤΥ꥽˴ޤ
	 */
	public void destroy() {
		super.destroy();
		removeAll();
		hawkContext.dispose();
	}
	
	/**
	 * ץåȤμ¹Ԥ򳫻Ϥޤ
	 */
	public void start() {
		super.start();
		hawkContext.hawkPanel.requestFocus();
	}
	
	/**
	 * ץåȤμ¹Ԥߤޤ
	 */
	/*
	public void stop() {
		super.stop();
	}
	*/
	
	/**
	 * 륢ץåȤˤĤƤξ֤ޤ
	 * 
	 * @return    륢ץåȤξ
	 */
	public String getAppletInfo() {
		return Resource.getMessage("kagetaka.copyright", null);
	}
	
	/**
	 * 륢ץåȤǤѥ᡼֤ͤޤ
	 * 
	 * @return    륢ץåȤǤѥ᡼
	 */
	public String[][] getParameterInfo() {
		return new String[][] {
			{"url"       , "URL"    , "read file url."                           },
			{"font"      , "string" , "java's font string(fontname-style-size)." },
			{"parser"    , "class"  , "parser class."                            },
			{"addressbar", "boolean", "use addressbar."                          },
			{"scrollbar" , "boolean", "use scrllbar."                            },
			{"spingif"   , "boolean", "spin gif image."                          },
			{"spinjpeg"  , "boolean", "spin jpeg image."                         },
			{"loglevel"  , "int"    , "log level."                               },
		};
	}
	
	/**
	 * ֥ѥͥκԤޤ
	 * 
	 * @param     g Graphics ɥ
	 */
	public void update(Graphics g)
	{
		paint(g);
	}
	
//### Private
	/** ǡɽޤ */
	private void open(URL url)
	{
		hawkContext.load(new Request(url, null, getDocumentBase(), null, false, false, Request.CACHE_NORMAL));
	}
	
	/** ꥽ʸ */
	private String getMessage(String key, String[] args)
	{
		return Resource.getMessage(RESOURCE, key, args);
	}
	
//### HawkAppletContext
	/** 륦ɥѤα륳ƥ */
	private class HawkAppletContext
		implements HawkContext,PanelListener,EngineListener
	{
		private AppletReporter reporter = new AppletReporter();
		
		private HawkEngine     hawkEngine;
		private HawkPanel      hawkPanel;
		private RequestList    requestList;
		private boolean        active;
		
		// ꥯȡ
		private Request viewRequest;            // ɽΥڡΥꥯ
		private Request nextRequest;            // ɤ߹Υꥯ
		
		/** 󥹥󥹤 */
		private HawkAppletContext()
		{
			hawkEngine = new HawkEngine (this);
			hawkPanel  = new HawkPanel  (this,
			                             new BasicScrollbar(BasicScrollbar.HORIZONTAL),
			                             new BasicScrollbar(BasicScrollbar.VERTICAL  ));
			requestList= new RequestList(    );
			
			hawkEngine.setEngineListener(this);
			hawkPanel .setPanelListener (this);
			
			// ˥塼ѹ
			initContextMenu();
			
			//### TODO Ȥꤢ
			active = true;
		}
		
		/** ѥͥΥƥȥ˥塼 */
		private void initContextMenu()
		{
			String   label = getMessage("contextmenu.openbrowser", null);
			Vector   items = null;
			MenuItem mi    = null;
			
			mi = new MenuItem(label);
			mi.addActionListener(new MenuActionListener(MenuActionListener.MAIN));
			items = hawkPanel.getMainMenuItems();
			items.addElement(new MenuItem("-"));
			items.addElement(mi);
			
			mi = new MenuItem(getMessage("contextmenu.dumpstatus", null));
			mi.addActionListener(new MenuActionListener(MenuActionListener.STATUS));
			items = hawkPanel.getMainMenuItems();
			items.addElement(new MenuItem("-"));
			items.addElement(mi);
			
			mi = new MenuItem(label);
			mi.addActionListener(new MenuActionListener(MenuActionListener.LINK));
			items = hawkPanel.getLinkMenuItems();
			items.addElement(new MenuItem("-"));
			items.addElement(mi);
		}
		
		/** URL  */
		private synchronized void load(Request request)
		{
			hawkEngine.stop();
			// ɤ߹ߤξǡPOST ǡ¸ߤ϶ɽ
			if (request.getDocument() != null &&    // ɤ߹߻ˤ Document ͭ
			    request.postData      != null) {
				synchronized (this) {
					changeHistory(request);
				}
				setupCanvas(request);
				setStatus  (getMessage("error.dorepost.stop", null));
				return;
			}
			
			request.setMoved(false);
			
			synchronized (this) {
				// ƱڡΥե󥹤ˤưǺɤ߹ߤʤ
				if (viewRequest != null                      &&
				    viewRequest != request                   &&        // ɤ߹ߤǤϤʤ
				    request.useCache >= Request.CACHE_NORMAL &&        // åѲ
				    //### TODO POST ξ϶Ū˺ɤ߹ߡ֤ͭޤǤλ
				    request.postData == null                 &&
				    StringUtils.sameFile(viewRequest.getDocument().url, request.url)) {
					request.copyFrom(viewRequest);
					changeHistory(request);
					movePosition (true   );
					setStatus    (null   );
					return;
				}
//System.out.println("URL=["+url+"]");
				nextRequest = request;
			}
			
			hawkPanel.setCursor(Cursor.WAIT_CURSOR);
			hawkEngine.load(request);
		}
		
		/** ˴ */
		void dispose()
		{
			hawkEngine.dispose();
		}
		
		//### HawkContext
		/** 륦ɥѤα륳ƥȤ */
		public Reporter getReporter()
		{
			return reporter;
		}
		
		/** ġ륭åȤ֤ */
		public Toolkit getToolkit()
		{
			return hawkPanel.getToolkit();
		}
		
		/** ꤷ URL ɽ */
		public void openHawk(Request request)
		{
			if (supportProtocol.containsKey(request.url.getProtocol())) {
				load(request);
			} else {
				// ݡȤƤʤץȥϥ֥饦ǳ
				if (request.isNewWindow)
					getAppletContext().showDocument(request.url, "_blank");
				else
					getAppletContext().showDocument(request.url);
			}
		}
		
		/** ꤵ줿 URL Υ֤ͥ */
		public URLConnection getURLConnection(Request request)
			throws IOException, InterruptedException
		{
			URLConnection connection = request.url.openConnection();
			
			// å
			if (cookieManager != null) {
				String cookie = cookieManager.getCookie(request.url);
				if (cookie != null)
					connection.setRequestProperty("Cookie", cookie);
			}
			
			// POST ǡ³
			if (request.postData != null && connection instanceof HttpURLConnection)
				request.postData.send((HttpURLConnection)connection);
			
			// ³
			try {
				connection.connect();
			} finally {
			if (Thread.interrupted())
				throw new InterruptedException("connect");
			}
			
			// å
			if (cookieManager != null) {
				String key;
				int    index = 1;
				
				for (;;) {
					if ((key = connection.getHeaderFieldKey(index)) == null)
						break;
					
					key = key.toLowerCase();
					if (key.compareTo("set-cookie") == 0)
						setCookie(connection.getHeaderField(index), request.url);
					
					index++;
				}
			}
			
			return connection;
		}
		
		/**  ꤵ줿ɥȤ */
		public void download(Document document, URLConnection connection)
		{
			getAppletContext().showDocument(document.url);
		}
		
		/** ᡼֤ޤ */
		public Image createImage(ImageProducer producer)
		{
			return hawkPanel.createImage(producer);
		}
		
		/** ᡼֤ޤ */
		public Image createImage(int width, int height)
		{
			return hawkPanel.createImage(width, height);
		}
		
		/** ǥȥå֤ */
		public MediaTracker getMediaTracker()
		{
			return hawkPanel.getMediaTracker();
		}
		
		/** ꡼󥤥᡼ɲ */
		public void addImage(Image image, int width, int height)
		{
			hawkPanel.addImage(image, width, height);
			synchronized (this) {
				movePosition(false);
			}
		}
		
		/** եॢƥɲ */
		public void addFormItem(FormItem item)
		{
			hawkPanel.addFormItem(item);
		}
		
		/** ѥ򥻥åȥå */
		public void setupCanvas(Request request)
		{
			hawkPanel.setupCanvas(request);
		}
		
		/** ѥϰϤꤷƺɽ */
		public void repaint(int x, int y, int width, int height)
		{
			hawkPanel.repaintCanvas(x, y, width, height);
		}
		
		/** ơå */
		public void setStatus(String status)
		{
			if (active)
				showStatus(status);
			statusText = status;
		}
		
		/** Ūʥơå */
		public void setTemporaryStatus(String status)
		{
			if (active) {
				if (status != null)
					showStatus(status);
				else
					showStatus(statusText);
			}
		}
		
		/** ӥ塼ݡȤθߤΥ֤ */
		public Dimension getViewportSize()
		{
			return hawkPanel.getViewportSize();
		}
		
		/** ץ֤ */
		public Option getOption()
		{
			return option;
		}
		
		/** ꤷ URL ˬ䤷Ȥ뤫ɤ֤ */
		public boolean haveEverVisited(URL url)
		{
			return visitedUrl.containsKey(url.toString());
		}
		
		/** å */
		public void setCookie(String value, URL url)
		{
			if (cookieManager != null)
				cookieManager.setCookie(value, url);
		}
		
		//### PanelListener
		/**  */
		public void back()
		{
			Request request = (viewRequest != null ? viewRequest.getPrevious() : null);
			if (request != null)
				load(request);
		}
		
		/** 뤫ɤ */
		public boolean canBack()
		{
			return (viewRequest != null && viewRequest.getPrevious() != null);
		}
		
		/** ؿʤ */
		public void forward()
		{
			Request request = (viewRequest != null ? viewRequest.getNext() : null);
			if (request != null)
				load(request);
		}
		
		/** ؿʤ뤫ɤ */
		public boolean canForward()
		{
			return (viewRequest != null && viewRequest.getNext() != null);
		}
		
		/** ɹ */
		public void reload(boolean force)
		{
			if (viewRequest != null)
				load(viewRequest);
		}
		
		/** ɹǤ뤫ɤ */
		public boolean canReload()
		{
			return (viewRequest != null);
		}
		
		/**  */
		public void stop()
		{
			hawkEngine.stop();
		}
		
		/** ߤǤ뤫ɤ */
		public boolean canStop()
		{
			return hawkEngine.isLoading();
		}
		
		//### EngineListener
		/** ³Ϥ˸ƤӽФ */
		public void connecting(Request request)
		{
/*
			if (nextRequest != request)
				return;
*/
		}
		
		/** ³λȸƤӽФ */
		public void connected(Request request, boolean noerror)
		{
/*
			if (nextRequest != request)
				return;
*/
		}
		
		/** 褬åפ줿Ȥ˸ƤӽФ */
		public synchronized void renderingSkipped(Request request)
		{
			if (nextRequest != request)
				return;
			
			nextRequest = null;
			hawkPanel.setCursor(Cursor.DEFAULT_CURSOR);
		}
		
		/** 褬Ϥ˸ƤӽФ */
		public synchronized void renderingStarted(Request request)
		{
			if (nextRequest != request)
				return;
			
			changeHistory(request);
			hawkPanel.requestFocus();
		}
		
		/** 褬λȸƤӽФ */
		public synchronized void renderingStopped(Request request, boolean noerror)
		{
			if (nextRequest != request)
				return;
			
			renderingSkipped(request);
			movePosition(true);
		}
		
		/** ꤷ URL ɽ */
		public synchronized void refresh(Request baseRequest, Request newRequest)
		{
			if (viewRequest != baseRequest)
				return;
			
			load(newRequest);
		}
		
		//### private
		/** ѹ */
		private void changeHistory(Request request)
		{
			// ߤɽ֤¸
			if (viewRequest != null)
				viewRequest.setPosition(hawkPanel.getScrollPosition());
			
			// ξ
			if (!request.isMember()) {
				requestList.append(viewRequest, request);
				visitedUrl.put(request.url.toString(), "");
			}
			
			viewRequest = request;
			
			if (active)
				addressBar.setText(request.getDisplayURL());
		}
		
		/** ݥư */
		private void movePosition(boolean force)
		{
			if (viewRequest.isMoved())
				return;
			
			Point position;
			
			// 桼ư֤
			if ((position = viewRequest.getPosition()) != null) {
				// 桼ˤꥯ뤵Ƥʤ
				Point scpos = hawkPanel.getScrollPosition();
				if ((force || (scpos.x == 0 && scpos.y == 0)) &&
				    !hawkPanel.setScrollPosition(position.x, position.y, false))
					return;
			} else
			
			// # ե󥹤ΰ֤
			if (!hawkPanel.setScrollPosition(viewRequest.url.getRef(), force))
				return;
			
			viewRequest.setMoved(true);
		}
	}
	
//### AppletReporter
	/** 륦ɥѤΥݡ */
	private class AppletReporter
		implements Reporter
	{
		/** 󥹥󥹤 */
		private AppletReporter()
		{
		}
		
		/** 륦ɥѤΥݡ */
		public void report(int level, String message, int line, int column)
		{
			if (level < logLevel)
				return;
			
			StringBuffer sb = new StringBuffer();
			sb.append("#");
			
			switch (level) {
			case Reporter.ERROR  : sb.append("ERROR"  ); break;
			case Reporter.WARNING: sb.append("WARNING"); break;
			default: sb.append("MESSAGE");
			}
			
			sb.append(":");
			sb.append(line);
			sb.append(":");
			sb.append(column);
			sb.append(":");
			sb.append(message);
			System.out.println(sb.toString());
		}
		
		/** Υå٥֤ */
		public int getLevel()
		{
			return logLevel;
		}
	}
	
//### MenuActionListener
	/** ץꥱ󤫤鳫ѤΥꥹ */
	private class MenuActionListener
		implements ActionListener
	{
		private static final int MAIN   = 0;
		private static final int LINK   = 1;
		private static final int STATUS = 2;
		
		private int type;
		
		/** ꥹʤ */
		private MenuActionListener(int type)
		{
			this.type = type;
		}
		
		/** ϥɥ */
		public void actionPerformed(ActionEvent e)
		{
			switch (type) {
			case MAIN:
			case LINK:
				{
					URL url = (type == MAIN ? (hawkContext.viewRequest != null ? hawkContext.viewRequest.url : null)
					                        : hawkContext.hawkPanel.getSelectedLink());
					if (url == null)
						return;
					
					getAppletContext().showDocument(url);
				}
				break;
			case STATUS:
				System.out.println(Resource.getMessage("kagetaka.name.en"  , null) + " " +
				                   Resource.getMessage("kagetaka.version"  , null) + " " +
				                   "(Build: " + Resource.getMessage(BUILD_RESOURCE, "build.number", null) + " " +
				                                Resource.getMessage(BUILD_RESOURCE, "build.date"  , null) + ")");
				System.out.println(Resource.getMessage("kagetaka.copyright", null) + " " +
				                   Resource.getMessage("kagetaka.vender.en", null));
				break;
			}
		}
	}
}
