/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.protocol.https;

import net.hizlab.kagetaka.protocol.ProxySupported;

import java.io.InputStream;
import java.io.FilterInputStream;
import java.io.IOException;
import java.net.URL;
import java.util.StringTokenizer;
import java.util.Map;

import javax.net.ssl.SSLSocketFactory;
import javax.net.ssl.HostnameVerifier;
import javax.net.ssl.HttpsURLConnection;

import sun.net.ProgressData;
import sun.net.ProgressEntry;
import sun.net.www.MessageHeader;
import sun.net.www.protocol.https.AbstractDelegateHttpsURLConnection;

/**
 * å򥵥ݡȤ <code>https</code> ץȥΥͥɽޤ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.1 $
 */
public class SunDelegateHttpsURLConnection
	extends AbstractDelegateHttpsURLConnection
	implements ProxySupported
{
	private static final String HTTP_VERSION = "HTTP/1.1";
	
	private HttpsURLConnection httpsURLConnection;
	private String             proxyHost;
	private int                proxyPort;
	private String             proxyAuthorization;
	
	private MessageHeader responses;
	private InputStream   inputStream;
	
	/** 󥹥󥹤 */
	SunDelegateHttpsURLConnection(URL url,
	                              sun.net.www.protocol.http.Handler handler, 
	                              HttpsURLConnection httpsURLConnection,
	                              String proxyHost, int proxyPort)
		throws IOException
	{
		super(url, handler);
		this.httpsURLConnection = httpsURLConnection;
		this.proxyHost          = proxyHost;
		this.proxyPort          = proxyPort;
	}
	
	/** SSL åȥեȥ */
	protected SSLSocketFactory getSSLSocketFactory()
	{
		return httpsURLConnection.getSSLSocketFactory();
	}
	
	/** ۥ̾٥ե */
	protected HostnameVerifier getHostnameVerifier()
	{
		return httpsURLConnection.getHostnameVerifier();
	}
	
	/** ³ */
	public void connect()
		throws IOException
	{
		connect(true);
	}
	
	/** ³ */
	private void connect(boolean useCache)
		throws IOException
	{
		if (connected)
			return;
		
		if (proxyHost == null) {
			plainConnect();
		} else {
			proxiedConnect(url, proxyHost, proxyPort, useCache);
			if (!http.isCachedConnection())
				doTunneling();
		}
		
		if (inputStream == null)
			((SunHttpsClient)http).afterConnect();
	}
	
	/** ľ³ */
	protected void plainConnect()
		throws IOException
	{
		if (connected)
			return;
		
		http = SunHttpsClient.New(getSSLSocketFactory(),
		                          url,
		                          getHostnameVerifier());
		connected = true;
	}
	
	/** ץͳ³ */
	protected void proxiedConnect(URL url, String proxyHost, int proxyPort,boolean useCache)
		throws IOException
	{
		if (connected) 
			return;
		
		SecurityManager security = System.getSecurityManager(); 
		if (security != null) { 
			security.checkConnect(proxyHost, proxyPort); 
		}
		http = SunHttpsClient.New(getSSLSocketFactory(),
		                          url,
		                          getHostnameVerifier(),
		                          proxyHost, proxyPort, useCache);
		connected = true;
	}
	
	/** 饤Ȥ */
	public void setNewClient(URL url, boolean useCache)
		throws IOException
	{
		http = SunHttpsClient.New(getSSLSocketFactory(),
		                          url,
		                          getHostnameVerifier(),
		                          proxyHost, proxyPort, useCache);
		
		if (inputStream == null)
			((SunHttpsClient)http).afterConnect();
	}
	
	/** ץͥ */
	public void setProxiedClient(URL url, String proxyHost, int proxyPort, boolean useCache)
		throws IOException
	{
		proxiedConnect(url, proxyHost, proxyPort, useCache);
		if (!http.isCachedConnection())
			doTunneling();
		((SunHttpsClient)http).afterConnect();
	}
	
	/** ȥͥ */
	protected synchronized void doTunneling()
		throws IOException
	{
		MessageHeader responses = new MessageHeader();
		
		if (!connected)
			proxiedConnect(url, proxyHost, proxyPort, false);
		
		sendCONNECTRequest();
		http.parseHTTP(responses, new ProgressEntry(url.getFile(), null));
		
		StringTokenizer st = new StringTokenizer(responses.getValue(0));
		st.nextToken();
		
		int respCode = Integer.parseInt(st.nextToken().trim());
		
		// HTTP 200 OK
		if (respCode == HTTP_OK)
			return;
		
		this.responseCode = respCode;
		this.responses    = responses;
		this.inputStream  = new HttpInputStream(http.getInputStream());
	}
	
	/** CONNECT  */
	private void sendCONNECTRequest()
		throws IOException
	{
		MessageHeader requests = new MessageHeader();
		
		int port = url.getPort();
		if (port == -1)
		    port = url.getDefaultPort();
		
		requests.prepend("CONNECT " + url.getHost() + ":" + port + " " + HTTP_VERSION, null);
		
		String host = url.getHost();
		if (port != -1 && port != 80)
		    host += ":" + String.valueOf(port);
		requests.set("Host", host);
		
		copyRequestProperty(requests, "User-Agent"     );
		copyRequestProperty(requests, "Accept"         );
		copyRequestProperty(requests, "Accept-Language");
		copyRequestProperty(requests, "Accept-Encoding");
		copyRequestProperty(requests, "Accept-Charset" );
		
		if (proxyAuthorization != null)
			requests.set("Proxy-Authorization", proxyAuthorization);
		
		http.writeRequests(requests, null);
	}
	
	/** ꤵƤ */
	private void copyRequestProperty(MessageHeader requests, String key)
	{
		String value = getRequestProperty(key);
		if (value != null)
			requests.set(key, value);
	}
	
	/** ꥯȥץѥƥ */
	public void setRequestProperty(String key, String value)
	{
		if (key.equalsIgnoreCase("proxy-authorization"))
			this.proxyAuthorization = value;
		else
			super.setRequestProperty(key, value);
	}
	
	/** ץåȥȥ꡼ */
	public synchronized InputStream getInputStream()
		throws IOException
	{
		if (inputStream != null)
			return inputStream;
		
		return super.getInputStream();
	}
	
	/** 顼ξΥץåȥȥ꡼ */
	public InputStream getErrorStream()
	{
		if (connected && responseCode >= 400 && inputStream != null)
			return inputStream;
		
		return super.getErrorStream();
	}
	
	/** 쥹ݥ󥹥إå */
	public String getHeaderField(String name)
	{
		try {
			getInputStream();
		} catch (IOException e) {}
		
		if (responses != null)
			return responses.findValue(name);
		
		return super.getHeaderField(name);
	}
	
	/** 쥹ݥ󥹥إå */
	public Map getHeaderFields()
	{
		try {
			getInputStream();
		} catch (IOException e) {}
		
		if (responses != null)
			return responses.getHeaders();
		
		return super.getHeaderFields();
	}
	
	/** 쥹ݥ󥹥إå */
	public String getHeaderField(int n)
	{
		try {
			getInputStream();
		} catch (IOException e) {}
		
		if (responses != null)
			return responses.getValue(n);
		
		return super.getHeaderField(n);
	}
	
	/** 쥹ݥ󥹥إåΥ */
	public String getHeaderFieldKey(int n)
	{
		try {
			getInputStream();
		} catch (IOException e) {}
		if (responses != null)
			return responses.getKey(n);
		return super.getHeaderFieldKey(n);
	}
	
//### HttpInputStream
	/** ץåȥȥ꡼ */
	private class HttpInputStream
		extends FilterInputStream
	{
		/** 󥹥󥹤 */
		private HttpInputStream(InputStream is)
		{
			super(is);
		}
		
		public void close()
			throws IOException
		{
			try {
				super.close();
			} finally {
				SunDelegateHttpsURLConnection.this.http = null;
			} 
		}
	}
}
