/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.protocol.https;

import net.hizlab.kagetaka.Resource;
import net.hizlab.kagetaka.protocol.ProxySupported;
import net.hizlab.kagetaka.protocol.URLConnectionCache;

import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.net.ProtocolException;
import java.security.Permission;
import java.security.cert.Certificate;
import java.util.Map;

import javax.net.ssl.HttpsURLConnection;
import javax.net.ssl.SSLPeerUnverifiedException;
import javax.security.cert.X509Certificate;

/**
 * å򥵥ݡȤ <code>https</code> ץȥΥͥɽޤ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.1 $
 */
public class SunHttpsURLConnection
	extends HttpsURLConnection
	implements ProxySupported
{
	private SunDelegateHttpsURLConnection delegate;
	private URLConnectionCache.Entry      cacheEntry;
	
	/** 󥹥󥹤 */
	SunHttpsURLConnection(URL url, SunHandler handler, String proxyHost, int proxyPort,
	                      URLConnectionCache.Entry cacheEntry)
		throws IOException
	{
		super(url);
		
		this.delegate  = new SunDelegateHttpsURLConnection(url, handler, this, proxyHost, proxyPort);
		
		// إå
		cacheEntry.setupURLConnection(delegate);
		
		this.cacheEntry = cacheEntry;
	}
	
//### URLConnection
	public void         connect                   (                            ) throws IOException       {        delegate.connect                   (                    ); }
	public URL          getURL                    (                            )                          { return delegate.getURL                    (                    ); }
	public int          getContentLength          (                            )                          { return delegate.getContentLength          (                    ); }
	public String       getContentType            (                            )                          { return delegate.getContentType            (                    ); }
	public String       getContentEncoding        (                            )                          { return delegate.getContentEncoding        (                    ); }
	public long         getExpiration             (                            )                          { return delegate.getExpiration             (                    ); }
	public long         getDate                   (                            )                          { return delegate.getDate                   (                    ); }
	public long         getLastModified           (                            )                          { return delegate.getLastModified           (                    ); }
	public String       getHeaderField            (String name                 )                          { return delegate.getHeaderField            (name                ); }
	public int          getHeaderFieldInt         (String name, int  Default   )                          { return delegate.getHeaderFieldInt         (name, Default       ); }
	public long         getHeaderFieldDate        (String name, long Default   )                          { return delegate.getHeaderFieldDate        (name, Default       ); }
	public String       getHeaderFieldKey         (int n                       )                          { return delegate.getHeaderFieldKey         (n                   ); }
	public String       getHeaderField            (int n                       )                          { return delegate.getHeaderField            (n                   ); }
	public Object       getContent                (                            ) throws IOException       { return delegate.getContent                (                    ); }
	public InputStream  getInputStream            (                            ) throws IOException       { return delegate.getInputStream            (                    ); }
	public OutputStream getOutputStream           (                            ) throws IOException       { return delegate.getOutputStream           (                    ); }
	public String       toString                  (                            )                          { return delegate.toString                  (                    ); }
	public void         setDoInput                (boolean doinput             )                          {        delegate.setDoInput                (doinput             ); }
	public boolean      getDoInput                (                            )                          { return delegate.getDoInput                (                    ); }
	public void         setDoOutput               (boolean dooutput            )                          {        delegate.setDoOutput               (dooutput            ); }
	public boolean      getDoOutput               (                            )                          { return delegate.getDoOutput               (                    ); }
	public void         setAllowUserInteraction   (boolean allowuserinteraction)                          {        delegate.setAllowUserInteraction   (allowuserinteraction); }
	public boolean      getAllowUserInteraction   (                            )                          { return delegate.getAllowUserInteraction   (                    ); }
	public void         setUseCaches              (boolean usecaches           )                          {        delegate.setUseCaches              (usecaches           ); }
	public boolean      getUseCaches              (                            )                          { return delegate.getUseCaches              (                    ); }
	public void         setIfModifiedSince        (long ifmodifiedsince        )                          {        delegate.setIfModifiedSince        (ifmodifiedsince     ); }
	public long         getIfModifiedSince        (                            )                          { return delegate.getIfModifiedSince        (                    ); }
	public boolean      getDefaultUseCaches       (                            )                          { return delegate.getDefaultUseCaches       (                    ); }
	public void         setDefaultUseCaches       (boolean defaultusecaches    )                          {        delegate.setDefaultUseCaches       (defaultusecaches    ); }
	public void         setRequestProperty        (String key, String value    )
	{
		delegate  .setRequestProperty(key, value);
		cacheEntry.addRequestProperty(key, value);
	}
	public String       getRequestProperty        (String key                  )                          { return delegate.getRequestProperty        (key                 ); }
	public void         addRequestProperty        (String key, String value    )                          {        delegate.addRequestProperty        (key, value          ); }
	public Map          getRequestProperties      (                            )                          { return delegate.getRequestProperties      (                    ); }
	public void         setInstanceFollowRedirects(boolean shouldFollow        )                          {        delegate.setInstanceFollowRedirects(shouldFollow        ); }
	public boolean      getInstanceFollowRedirects(                            )                          { return delegate.getInstanceFollowRedirects(                    ); }
	public Permission   getPermission             (                            ) throws IOException       { return delegate.getPermission             (                    ); }
	
//### HttpURLConnection
	public void         setRequestMethod          (String method               ) throws ProtocolException {        delegate.setRequestMethod          (method              ); }
	public String       getRequestMethod          (                            )                          { return delegate.getRequestMethod          (                    ); }
	public int          getResponseCode           (                            ) throws IOException       { return delegate.getResponseCode           (                    ); }
	public String       getResponseMessage        (                            ) throws IOException       { return delegate.getResponseMessage        (                    ); }
	public void         disconnect                (                            )                          {        delegate.disconnect                (                    ); }
	public boolean      usingProxy                (                            )                          { return delegate.usingProxy                (                    ); }
	public InputStream  getErrorStream            (                            )                          { return delegate.getErrorStream            (                    ); }
	
//### Object
	protected void    finalize(          ) throws Throwable {        delegate.finalize(   ); }
	public    boolean equals  (Object obj)                  { return delegate.equals  (obj); }
	public    int     hashCode(          )                  { return delegate.hashCode(   ); }
	
//### sun.net.www.protocol.http.HttpURLConnection
	protected void setNewClient    (URL url                                                   ) throws IOException { delegate.setNewClient    (url,                       false   ); }
	protected void setNewClient    (URL url,                                  boolean useCache) throws IOException { delegate.setNewClient    (url,                       useCache); }
	protected void setProxiedClient(URL url, String proxyHost, int proxyPort                  ) throws IOException { delegate.setProxiedClient(url, proxyHost, proxyPort          ); }
	protected void setProxiedClient(URL url, String proxyHost, int proxyPort, boolean useCache) throws IOException { delegate.setProxiedClient(url, proxyHost, proxyPort, useCache); }
	
//### sun.net.www.protocol.https.AbstractDelegateHttpsURLConnection
	protected boolean isConnected (            ) { return delegate.isConnected (    ); }
	protected void    setConnected(boolean conn) {        delegate.setConnected(conn); }
	public X509Certificate[] getServerCertificateChain()
	{
		try {
			return delegate.getServerCertificateChain();
		} catch (SSLPeerUnverifiedException e) {
			return null;
		}
	}
	
//### javax.net.ssl.HttpsURLConnection
	public String         getCipherSuite       ()                                   { return delegate.getCipherSuite       (); }
	public Certificate [] getLocalCertificates ()                                   { return delegate.getLocalCertificates (); }
	public Certificate [] getServerCertificates() throws SSLPeerUnverifiedException { return delegate.getServerCertificates(); }
}
