/* ----- BEGIN LICENSE BLOCK -----
 * Version: MPL 1.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is the Kagetaka Libraries.
 *
 * The Initial Developer of the Original Code is Hizuya Atsuzaki
 * Portions created by the Initial Developer are Copyright (C) 2003
 * the Initial Developer. All Rights Reserved.
 *
 * Contributor(s): Hizuya Atsuzaki <hizuya@hizlab.net>
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either the GNU General Public License Version 2 or later (the "GPL"), or
 * the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 *
 * ----- END LICENSE BLOCK ----- */
package net.hizlab.kagetaka.rendering;

import java.awt.Point;
import java.net.URL;

/**
 * ɤ߹ߥꥯȤɽ饹Ǥ
 * 
 * @author  <A HREF="mailto:hizuya@hizlab.net">Hizuya Atsuzaki</A>
 * @version $Revision: 1.2 $
 */
public class Request
{
	/** ɬåȤ */
	public static final int CACHE_MUST   = 3;
	/** ξϥåȤڤʤåԤ */
	public static final int CACHE_NORMAL = 2;
	/** ɬåԤ */
	public static final int CACHE_CHECK  = 1;
	/** åȤʤ */
	public static final int CACHE_NONE   = 0;
	
	/** ꥯ URL */
	public final URL      url;
	/** POST ǡ */
	public final PostData postData;
	/** 󥯸 URL */
	public final URL      referer;
	/** å */
	public final String   target;
	/** ɥǳɤ */
	public final boolean  isNewWindow;
	/** ĥ⡼ɤǳɤ */
	public final boolean  isExtension;
	/** åѲǽɤ */
	public final int      useCache;
	
	// 
	Request next;
	Request prev;
	
	private Document document;
	private boolean  isMoved;
	private Point    position;
	
	/**
	 * ꥯȤޤ
	 * 
	 * @param     url       URL
	 * @param     pd        POST Ϥǡ
	 *                      POST ǤϤʤ <code>null</code>
	 * @param     referer   󥯸 URL
	 *                      󥯤éäΤǤϤʤ <code>null</code>
	 * @param     target    å
	 *                      åȤꤵƤʤ <code>null</code>
	 * @param     newWindow ɥǳ <code>true</code>
	 *                      ʳξ <code>false</code>
	 * @param     extension ĥ⡼ɤǳ <code>true</code>
	 *                      ʳξ <code>false</code>
	 * @param     cache     åѥ
	 */
	public Request(URL url, PostData pd, URL referer, String target,
	               boolean newWindow, boolean extension, int cache)
	{
		this.url         = url;
		this.postData    = pd;
		this.referer     = referer;
		this.target      = target;
		this.isNewWindow = newWindow;
		this.isExtension = extension;
		this.useCache    = cache;
	}
	
	/**
	 * ɥȤ֤ޤ
	 * 
	 * @return    ɥ
	 */
	public Document getDocument()
	{
		return document;
	}
	
	/**
	 * ɥȤꤷޤ
	 * 
	 * @param     document ɥ
	 */
	void setDocument(Document document)
	{
		this.document = document;
	}
	
	/**
	 * ꥯȤηƤ򥳥ԡޤ
	 * 
	 * @param     src ԡΥꥯ
	 */
	public void copyFrom(Request src)
	{
		document = new Document(url, src.getDocument());
	}
	
	/**
	 * ꥯȤη̤ URL ֤ޤ
	 * 
	 * @return    ꥯȤη̤ URL
	 */
	public String getDisplayURL()
	{
		String ref = url.getRef();
		if (ref == null)
			return document.url.toString();
		
		String urlString = document.url.toString();
		int p = urlString.indexOf('#');
		if (p != -1)
			return urlString.substring(0, p + 1) + ref;
		
		return urlString + "#" + ref;
	}
	
	/**
	 * ݥưѤߤɤ֤ޤ
	 * 
	 * @return    ݥ󤬰ưѤߤξ <code>true</code>
	 *            ʳξ <code>false</code>
	 */
	public boolean isMoved()
	{
		return isMoved;
	}
	
	/**
	 * ݥưѤߤɤꤷޤ
	 * 
	 * @param     moved ݥ󤬰ưѤߤξ <code>true</code>
	 *                  ʳξ <code>false</code>
	 */
	public void setMoved(boolean moved)
	{
		this.isMoved = moved;
	}
	
	/**
	 * ݥ֤ޤ
	 * 
	 * @return    ݥ
	 */
	public Point getPosition()
	{
		return position;
	}
	
	/**
	 * ݥꤷޤ
	 * 
	 * @param     position ݥ
	 */
	public void setPosition(Point position)
	{
		this.position = position;
	}
	
	/**
	 * μΥꥯȤ֤ޤ
	 * 
	 * @return    Υꥯ
	 */
	public Request getNext()
	{
		return next;
	}
	
	/**
	 * ΥꥯȤ֤ޤ
	 * 
	 * @return    Υꥯ
	 */
	public Request getPrevious()
	{
		return prev;
	}
	
	/**
	 * ΥФɤ֤ޤ
	 * 
	 * @return    ΥФξ <code>true</code>
	 *            ʳξ <code>false</code>
	 */
	public boolean isMember()
	{
		return (prev != null || next != null);
	}
}
